# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - EventAggregator Macro

    @copyright: 2008, 2009 by Paul Boddie <paul@boddie.org.uk>
    @copyright: 2000-2004 Juergen Hermann <jh@web.de>,
                2005-2008 MoinMoin:ThomasWaldmann.
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.Page import Page
from MoinMoin import wikiutil, search, version
import calendar
import re

try:
    set
except NameError:
    from sets import Set as set

__version__ = "0.1"

Dependencies = ['pages']

# Regular expressions where MoinMoin does not provide the required support.

category_regexp = None
definition_list_regexp = re.compile(ur'^\s+(?P<term>.*?)::\s(?P<desc>.*?)$', re.UNICODE | re.MULTILINE)
date_regexp = re.compile(ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})', re.UNICODE)
month_regexp = re.compile(ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})', re.UNICODE)

# Date labels.

month_labels = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]
weekday_labels = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

# Utility functions.

def isMoin15():
    return version.release.startswith("1.5.")

def getCategoryPattern(request):
    global category_regexp

    try:
        return request.cfg.cache.page_category_regexact
    except AttributeError:

        # Use regular expression from MoinMoin 1.7.1 otherwise.

        if category_regexp is None:
            category_regexp = re.compile(u'^%s$' % ur'(?P<all>Category(?P<key>(?!Template)\S+))', re.UNICODE)
        return category_regexp

# The main activity functions.

def getPages(pagename, request):

    "Return the links minus category links for 'pagename' using the 'request'."

    query = search.QueryParser().parse_query('category:%s' % pagename)
    if isMoin15():
        results = search.searchPages(request, query)
        results.sortByPagename()
    else:
        results = search.searchPages(request, query, "page_name")

    cat_pattern = getCategoryPattern(request)
    pages = []
    for page in results.hits:
        if not cat_pattern.match(page.page_name):
            pages.append(page)
    return pages

def getPrettyPageName(page):

    "Return a nicely formatted title/name for the given 'page'."

    return page.split_title(force=1).replace("_", " ").replace("/", u"  ")

def getEventDetails(page):

    "Return a dictionary of event details from the given 'page'."

    event_details = {}

    if page.pi["format"] == "wiki":
        for match in definition_list_regexp.finditer(page.body):
            # Permit case-insensitive list terms.
            term = match.group("term").lower()
            desc = match.group("desc")
            if term in ("start", "end"):
                desc = getDate(desc)
            if desc is not None:
                event_details[term] = desc

    return event_details

def getDate(s):

    "Parse the string 's', extracting and returning a date string."

    m = date_regexp.search(s)
    if m:
        return tuple(map(int, m.groups()))
    else:
        return None

def getMonth(s):

    "Parse the string 's', extracting and returning a month string."

    m = month_regexp.search(s)
    if m:
        return tuple(map(int, m.groups()))
    else:
        return None

def daterange(first, last):
    results = []

    months_only = len(first) == 2
    start_year = first[0]
    end_year = last[0]

    for year in range(start_year, end_year + 1):
        if year < end_year:
            end_month = 12
        else:
            end_month = last[1]

        if year > start_year:
            start_month = 1
        else:
            start_month = first[1]

        for month in range(start_month, end_month + 1):
            if months_only:
                results.append((year, month))
            else:
                if month < end_month:
                    _wd, end_day = calendar.monthrange(year, month)
                else:
                    end_day = last[2]

                if month > start_month:
                    start_day = 1
                else:
                    start_day = first[2]

                for day in range(start_day, end_day + 1):
                    results.append((year, month, day))

    return results

def getColour(s):
    colour = [0, 0, 0]
    digit = 0
    for c in s:
        colour[digit] += ord(c)
        colour[digit] = colour[digit] % 256
        digit += 1
        digit = digit % 3
    return tuple(colour)

def getBlackOrWhite(colour):
    if sum(colour) / 3.0 > 127:
        return (0, 0, 0)
    else:
        return (255, 255, 255)

def execute(macro, args):

    """
    Execute the 'macro' with the given 'args': an optional list of selected
    category names (categories whose pages are to be shown), together with
    optional named arguments of the following forms:

      start=YYYY-MM     shows event details starting from the specified month
      end=YYYY-MM       shows event details ending at the specified month

      mode=calendar     shows a calendar view of events
      mode=list         shows a list of events by month

      names=daily       shows the name of an event on every day of that event
      names=weekly      shows the name of an event once per week
    """

    request = macro.request
    fmt = macro.formatter
    page = fmt.page
    _ = request.getText

    # Interpret the arguments.

    try:
        parsed_args = args and wikiutil.parse_quoted_separated(args, name_value=False) or []
    except AttributeError:
        parsed_args = args.split(",")

    parsed_args = [arg for arg in parsed_args if arg]

    # Get special arguments.

    category_names = []
    calendar_start = None
    calendar_end = None
    mode = "calendar"
    name_usage = "daily"

    for arg in parsed_args:
        if arg.startswith("start="):
            calendar_start = getMonth(arg[6:])
        elif arg.startswith("end="):
            calendar_end = getMonth(arg[4:])
        elif arg.startswith("mode="):
            mode = arg[5:]
        elif arg.startswith("names="):
            name_usage = arg[6:]
        else:
            category_names.append(arg)

    # Generate a list of events found on pages belonging to the specified
    # categories, as found in the macro arguments.

    events = []
    shown_events = {}

    earliest = None
    latest = None

    for category_name in category_names:

        # Get the pages and page names in the category.

        pages_in_category = getPages(category_name, request)

        # Visit each page in the category.

        for page_in_category in pages_in_category:
            pagename = page_in_category.page_name

            # Get a real page, not a result page.

            real_page_in_category = Page(request, pagename)
            event_details = getEventDetails(real_page_in_category)

            # Define the event as the page together with its details.

            event = (real_page_in_category, event_details)
            events.append(event)

            # Test for the suitability of the event.

            if event_details.has_key("start") and event_details.has_key("end"):

                start_month = event_details["start"][:2]
                end_month = event_details["end"][:2]

                # Compare the months of the dates to the requested calendar
                # window, if any.

                if (calendar_start is None or end_month >= calendar_start) and \
                    (calendar_end is None or start_month <= calendar_end):

                    if earliest is None or start_month < earliest:
                        earliest = start_month
                    if latest is None or end_month > latest:
                        latest = end_month

                    # Store the event in the month-specific dictionary.

                    first = max(start_month, calendar_start or start_month)
                    last = min(end_month, calendar_end or end_month)

                    for event_month in daterange(first, last):
                        if not shown_events.has_key(event_month):
                            shown_events[event_month] = []
                        shown_events[event_month].append(event)

    # Make a calendar.

    output = []

    if mode == "list":
        output.append(fmt.bullet_list(on=1, attr={"class" : "event-listings"}))

    # Visit all months in the requested range, or across known events.

    first = calendar_start or earliest
    last = calendar_end or latest

    for year, month in daterange(first, last):

        # Either output a calendar view...

        if mode == "calendar":

            # Output a month.

            output.append(fmt.table(on=1, attrs={"tableclass" : "event-month"}))

            output.append(fmt.table_row(on=1))
            output.append(fmt.table_cell(on=1, attrs={"class" : "event-month-heading", "colspan" : "7"}))
            output.append(fmt.span(on=1))
            output.append(fmt.text(_(month_labels[month - 1]))) # zero-based labels
            output.append(fmt.span(on=0))
            output.append(fmt.text(" "))
            output.append(fmt.span(on=1))
            output.append(fmt.text(year))
            output.append(fmt.span(on=0))
            output.append(fmt.table_cell(on=0))
            output.append(fmt.table_row(on=0))

            # Weekday headings.

            output.append(fmt.table_row(on=1))

            for weekday in range(0, 7):
                output.append(fmt.table_cell(on=1, attrs={"class" : "event-weekday-heading"}))
                output.append(fmt.text(_(weekday_labels[weekday])))
                output.append(fmt.table_cell(on=0))

            output.append(fmt.table_row(on=0))

            # Process the days of the month.

            start_weekday, number_of_days = calendar.monthrange(year, month)

            # The start weekday is the weekday of day number 1.
            # Find the first day of the week, counting from below zero, if
            # necessary, in order to land on the first day of the month as
            # day number 1.

            first_day = 1 - start_weekday

            while first_day <= number_of_days:

                # Find events in this week and determine how to mark them on the
                # calendar.

                week_events = []
                week_coverage = set()

                week_start = (year, month, max(first_day, 1))
                week_end = (year, month, min(first_day + 6, number_of_days))

                # Get event details.

                for event in shown_events.get((year, month), []):
                    event_page, event_details = event

                    # Test for the event in the current week.

                    if event_details["start"] <= week_end and event_details["end"] >= week_start:

                        # Find the coverage of this week for the event.

                        event_start = max(event_details["start"], week_start)
                        event_end = min(event_details["end"], week_end)
                        event_coverage = set(daterange(event_start, event_end))

                        # Update the overall coverage.

                        week_coverage.update(event_coverage)

                        # Try and fit the event into the events list.

                        for i, (coverage, events) in enumerate(week_events):

                            # Where the event does not overlap with the current
                            # element, add it alongside existing events.

                            if not coverage.intersection(event_coverage):
                                events.append(event)
                                week_events[i] = coverage.union(event_coverage), events
                                break

                        # Make a new element in the list if the event cannot be
                        # marked alongside existing events.

                        else:
                            week_events.append((event_coverage, [event]))

                # Output a week, starting with the day numbers.

                output.append(fmt.table_row(on=1))

                for weekday in range(0, 7):
                    day = first_day + weekday
                    date = (year, month, day)

                    # Output out-of-month days.

                    if day < 1 or day > number_of_days:
                        output.append(fmt.table_cell(on=1, attrs={"class" : "event-day-heading event-day-excluded"}))
                        output.append(fmt.table_cell(on=0))

                    # Output normal days.

                    else:
                        if date in week_coverage:
                            output.append(fmt.table_cell(on=1, attrs={"class" : "event-day-heading event-day-busy"}))
                        else:
                            output.append(fmt.table_cell(on=1, attrs={"class" : "event-day-heading event-day-empty"}))

                        output.append(fmt.div(on=1))
                        output.append(fmt.span(on=1, css_class="event-day-number"))
                        output.append(fmt.text(day))
                        output.append(fmt.span(on=0))
                        output.append(fmt.div(on=0))

                        # End of day.

                        output.append(fmt.table_cell(on=0))

                # End of day numbers.

                output.append(fmt.table_row(on=0))

                # Visit each set of scheduled events.

                for coverage, events in week_events:

                    # Output each set.

                    output.append(fmt.table_row(on=1))

                    # Then, output day details.

                    for weekday in range(0, 7):
                        day = first_day + weekday
                        date = (year, month, day)

                        # Skip out-of-month days.

                        if day < 1 or day > number_of_days:
                            output.append(fmt.table_cell(on=1, attrs={"class" : "event-day event-day-excluded"}))
                            output.append(fmt.table_cell(on=0))
                            continue

                        # Output the day.

                        if date in coverage:
                            output.append(fmt.table_cell(on=1, attrs={"class" : "event-day event-day-busy"}))
                        else:
                            output.append(fmt.table_cell(on=1, attrs={"class" : "event-day event-day-empty"}))

                        # Get event details for the current day.

                        for event_page, event_details in events:
                            if not (event_details["start"] <= date <= event_details["end"]):
                                continue

                            # Get a pretty version of the page name.

                            pretty_pagename = getPrettyPageName(event_page)

                            # Generate a colour for the event.

                            bg = getColour(event_page.page_name)
                            fg = getBlackOrWhite(bg)

                            css_classes = ["event-summary"]

                            if event_details["start"] == date:
                                css_classes.append("event-starts")
                                start_of_event = 1
                            else:
                                start_of_event = 0

                            if event_details["end"] == date:
                                css_classes.append("event-ends")

                            # Output the event.

                            if name_usage == "daily" or start_of_event or weekday == 0 or day == 1:
                                hide_text = 0
                            else:
                                hide_text = 1

                            output.append(fmt.div(on=1, css_class=(" ".join(css_classes)),
                                style=("background-color: rgb(%d, %d, %d); color: rgb(%d, %d, %d);" % (bg + fg))))

                            if not hide_text:
                                output.append(event_page.link_to_raw(request, wikiutil.escape(pretty_pagename)))

                            output.append(fmt.div(on=0))

                        # End of day.

                        output.append(fmt.table_cell(on=0))

                    # End of set.

                    output.append(fmt.table_row(on=0))

                # Process the next week...

                first_day += 7

            # End of month.

            output.append(fmt.table(on=0))

        # Or output a summary view...

        elif mode == "list":

            output.append(fmt.listitem(on=1, attr={"class" : "event-listings-month"}))
            output.append(fmt.div(on=1, attr={"class" : "event-listings-month-heading"}))
            output.append(fmt.span(on=1))
            output.append(fmt.text(_(month_labels[month - 1]))) # zero-based labels
            output.append(fmt.span(on=0))
            output.append(fmt.text(" "))
            output.append(fmt.span(on=1))
            output.append(fmt.text(year))
            output.append(fmt.span(on=0))
            output.append(fmt.div(on=0))

            output.append(fmt.bullet_list(on=1, attr={"class" : "event-month-listings"}))

            for event_page, event_details in shown_events.get((year, month), []):

                # Get a pretty version of the page name.

                pretty_pagename = getPrettyPageName(event_page)

                output.append(fmt.listitem(on=1, attr={"class" : "event-listing"}))

                # Link to the page using the pretty name.

                output.append(event_page.link_to_raw(request, wikiutil.escape(pretty_pagename)))

                # Add the event details.

                output.append(fmt.definition_list(on=1, attr={"class" : "event-details"}))

                for key, value in event_details.items():
                    output.append(fmt.definition_term(on=1))
                    output.append(fmt.text(key))
                    output.append(fmt.definition_term(on=0))
                    output.append(fmt.definition_desc(on=1))
                    output.append(fmt.text(value))
                    output.append(fmt.definition_desc(on=0))

                output.append(fmt.definition_list(on=0))
                output.append(fmt.listitem(on=0))

            output.append(fmt.bullet_list(on=0))

    if mode == "list":
        output.append(fmt.bullet_list(on=0))

    return ''.join(output)

# vim: tabstop=4 expandtab shiftwidth=4
